#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <asm/io.h>
#include <time.h>
#include <sys/time.h>
#include <pthread.h>


/* Defaults */
#define PORT     0x378
#define PORT_IN  PORT+1
#define VERSION  "0.10"
#define APP_NAME "Crane"
#define BIN_NAME "crane"
#define MAX_PICS 5
#define DEFAULT_CONFIG_FILE ".crane.cfg"
#define DEFAULT_PIC_PATH    "default"


/* Macs */
#define IS_NUM(x) ( ((int)x>47) & ((int)x<58) )
#define CLAW_SHIFT(x, c) ( (c) ? x+1 : x )
#define TIMES_TO_TICKS(t) (t/100)


/* Flags and such */
static char g_config_file[50] = DEFAULT_CONFIG_FILE;
static int is_verbose   = 1;
static int is_take_pics = 1;
static long int xtime   = 0;
static long int ytime   = 0;
pthread_t cam_thread;


/* Demo */
static enum {STROKE, SWIRL, DANGLE} d_mode;


/* Pins */
#define MV_LEFT       24
#define MV_RIGHT      16
#define MV_FRONT      64
#define MV_BACK       96
#define MV_CLAW_UP    6
#define MV_CLAW_DOWN  4
#define MV_CLAW_CLOSE 1
static int pin_status[4];
static int pin_fb_home   = 0;
static int pin_lr_home   = 1;
static int pin_claw_up   = 2;
static int pin_claw_down = 3;


/* Fowards */
static int get_input(void);


/* Move by clock ticks */
static void move_tick(
	int direction,
	int status_idx,
	int ticks,
	int force)
{
	int i;
    clock_t ct_a;

	if(pin_status[status_idx] || force)
    {
        ct_a = clock();
        for(i=0; i<ticks; i++)
        {
            outb(direction, PORT);
            if(clock()-ct_a > ticks)
              break;
        }
	    outb(0, PORT);
    }
}


/* Duh */
static void claw_fetch(void)
{
    /* Drop */
    while(!pin_status[pin_claw_down])
    {
        get_input();
        outb(MV_CLAW_DOWN, PORT);
    }
    usleep(200);
    outb(0, PORT);
    sleep(1);

    /* Up */
    while(!pin_status[pin_claw_up])
    {
        get_input();
        outb(CLAW_SHIFT(MV_CLAW_UP, 1), PORT);
    }
    outb(CLAW_SHIFT(MV_CLAW_CLOSE, 0), PORT);
    sleep(1);
}


/* What the inputs are */
static void print_pins(void)
{
	int i;

	for(i=0; i<4; i++)
	  printf("pins[%d]: %d\n", i, pin_status[i]);
}


/* What is the crane tellin us? 
 * Return high 4 bits representing HOME and CLAW status 
 */
static int get_input(void)
{
        unsigned char ret;
        int i;
        
        ret = inb(PORT_IN);
        ret >>= 4;

        for(i=0; i<4; i++)
          pin_status[i] = (ret>>i)%2;

        return ret;
}


/* Go to the crane's origin (Where the prizes are dropped) */
static void go_origin(int is_claw)
{
    int is_good = 0;
    int output;

    /* Is the claw grabbin some booty */
    
    /* Dont ask */
    move_tick(CLAW_SHIFT(MV_LEFT, is_claw), 
              pin_lr_home, 100000, 1);
	move_tick(CLAW_SHIFT(MV_FRONT, is_claw), 
              pin_fb_home, 100000, 1);

    get_input();
    while(!is_good)
    {
        output = 0;

        /* Bring claw up */
        while(!pin_status[pin_claw_up])
        {
            outb(CLAW_SHIFT(MV_CLAW_UP, is_claw), PORT);
            get_input();
            
            /* Stop yo */
            if(pin_status[pin_claw_up])
              outb(0, PORT);
        }

        /* Claw working */
        output = CLAW_SHIFT(output, is_claw);
        
        /* Left right home */
        if(pin_status[pin_lr_home])
          output += MV_LEFT;
        
        /* Front back home */
        if(pin_status[pin_fb_home])
          output += MV_FRONT;

        /* Go */
        outb(output, PORT);

        if(is_claw)
          output--;

        if(output == 0)
          is_good = 1;

        /* Update */
        get_input();
		
#ifdef DEBUG
		print_pins();
#endif
    }
}


/* Take pics */
static void *get_pics(void *path)
{
    char cmd[128];

    //sprintf(cmd, "streamer -d -r 2 0:10 -o %s/%s.avi -f jpeg", 
    //(char *)path, tag);
    //snprintf(cmd, 128, "vgrabbj -f %s/%s.jpg", (char *)path, tag); 
    snprintf(cmd, 128, "/root/crane/videocapture.sh %s &", (char *)path); 
    
    system(cmd);

    /* For static pics 
     * for(i=0; i<MAX_PICS; i++)
    {
        sprintf(cmd, "vgrabbj -f %s/%d.jpg", (char *)path, i); 
        system(cmd);
    }
    */

    return;
}


/* Go to the point and get some prizes */
static void cranage(
    double x, 
    double y,
    char  *pic_path)
{
    /* Take pics in mega quasi-semi-parallel mode */
    if(is_take_pics)
      pthread_create(&cam_thread, NULL, get_pics, pic_path);
    
    /* go_point() */
    move_tick(MV_RIGHT, pin_lr_home, x, 1);
    move_tick(MV_BACK, pin_fb_home, y, 1);
    
    claw_fetch();
    go_origin(1);
    outb(0, PORT);
}


/* Load configs */
static void load_configs(char *fname)
{
    FILE *fp;
    char line[64];
    
    memset(line, '\0', sizeof(line));
    if((fp = fopen(fname, "r")))
    {
        /* Skip comments */
        while(fgets(line, 32, fp) && line[0] == ';');
     
        /* Tick times */
        xtime = TIMES_TO_TICKS(atol(strtok(line, ",")));
        ytime = TIMES_TO_TICKS(atol(strtok(NULL, ",")));
#ifdef DEBUG
        printf("loading xtime(%i) ytime(%i)\n", xtime, ytime);
#endif
        fclose(fp);
    }
    else /* dont die */
      printf("Error opening config file to load: %s\n",
              fname);
}


/* Save configs */
static void save_configs(
    char *fname,
    char *data)
{
    FILE *fp;
    
    if(!(fp = fopen(fname, "w")))
      printf("Error opening config file to save: %s\n",
              fname);
    fprintf(fp, data, sizeof(char) * (strlen(data) + 1));
    fclose(fp);
}
        

/* Initalize Crane */
static void init_crane(void)
{
    char save_data[256];
    char str[50];
	clock_t ct_a, ct_b;
    
    sprintf(save_data, "; Config created: %d\n", (unsigned int)time(NULL));

    if(is_verbose)
      printf("Initalizing crane...\n");

    /* Drop string */
    outb(MV_CLAW_DOWN, PORT);
    usleep(100000);
    outb(0, PORT);

    /* Raise string */
    while(!pin_status[pin_claw_up])
    {
        get_input();
        outb(CLAW_SHIFT(MV_CLAW_UP, 1), PORT);
    }
    outb(0, PORT);
    
    /* Go to your home */
    go_origin(0);

	/* Callibrate X-axis */
	pin_status[pin_lr_home] = 1;
	ct_a = clock();
	outb(MV_RIGHT, PORT);
	usleep(100000);
	while(pin_status[pin_lr_home])
  	  get_input();
	ct_b = clock();
	outb(0, PORT);
	xtime = ct_b-ct_a;

	if(is_verbose)
	  printf("X-Axis time: %d\n", (int)xtime);
	
	/* Callibrate Y-axis */
	pin_status[pin_fb_home] = 1;
	ct_a = clock();
	outb(MV_BACK, PORT);
	usleep(100000);
	while(pin_status[pin_fb_home])
  	  get_input();
	ct_b = clock();
	outb(0, PORT);
	ytime = ct_b-ct_a;

	if(is_verbose)
	  printf("Y-Axis time: %d\n", (int)ytime);

    /* Save to config file */
    sprintf(str, "%d,%d\n", (int)xtime, (int)ytime);
    strcat(save_data, str);
    save_configs(g_config_file, save_data);

    /* Return */
    go_origin(0);
}


/* Demo Crane */
static void demo_crane(void)
{
    int i;
    
    if(is_verbose)
      printf("Demoing crane...\n");

    /* Do some laps, have it lower claw,
     * and swing the claw so it taps the glass
     * possibly go for a pendulum effect
     * consider disco ball
     */
    d_mode = STROKE;
    switch(d_mode)
    {
        case STROKE:
            go_origin(0);
            move_tick(MV_RIGHT, pin_lr_home, (xtime*100)/2, 1);
            move_tick(MV_BACK, pin_lr_home, (ytime*100)/2, 1);
            for(i=0; i<(1+rand())%10; i++)
            {
                outb(MV_CLAW_DOWN, PORT);
                usleep(800000);
                get_input();
                while(!pin_status[pin_claw_up])
                {
                    get_input();
                    outb(CLAW_SHIFT(MV_CLAW_UP, 1), PORT);
                }
                usleep(300000);
            }
            outb(0, PORT);
            go_origin(0);
            break;

        case DANGLE:
            go_origin(0);
            move_tick(MV_RIGHT, pin_lr_home, (xtime*100)/2, 1);
            move_tick(MV_BACK, pin_lr_home, (ytime*100)/2, 1);
                
            outb(MV_CLAW_DOWN, PORT);
            usleep(300000);
            outb(0, PORT);
            usleep(600000);
            while(!pin_status[pin_claw_up])
            {
                get_input();
                outb(CLAW_SHIFT(MV_CLAW_UP, 1), PORT);
            }
            outb(0, PORT);
            go_origin(0);
            break;
        
        case SWIRL:
            break;
    }
}


/* Usage */
static void usage(void)
{
    printf("--{ %s -- v%s }--\n"
           "Usage: ./%s [-c <x,y>] <-e email> <-H hash> [-d] [-i] [-v] ... [-h]\n"
           "\t-c <x, y> Attack Coordinate (X, Y)\n"
           "\t-d Demo Mode\n"
           "\t-e User Email\n"
           "\t-H Hash for user's email\n"
           "\t-i Initalize Crane\n"
           "\t-n Do not take pictures\n"
           "\t-p Path for storing user's pictures.\n"
           "\t-h Help\n"
           "\t-q Quiet\n",
           APP_NAME, VERSION, BIN_NAME);
}


/* Main */
int main(
    int    argc,
    char **argv)
{
    int i;
    int call_crane = 0;
    long int x, y;
    char pic_path[256];
    char script_args[256];
    char email[32];
    char hash[64];

    srand(time(NULL));

    ioperm(PORT_IN, 1, 1);
    ioperm(PORT, 1, 1);
    
    /* Argz:
     * 
     * -i initalize
     * -e email
     * -H hash
     * -h help
     * -c <int, int>
     * -d demo
     * -n dont take pics
     * -p pic path
     * -q quiet
     */
    if(argc == 1)
    {
        usage();
        exit(-1);
    }
    else
      load_configs(g_config_file);

    for(i=1; i<argc; i++)
    {
        if(argv[i][0] != '-')
          continue;
          
        switch(argv[i][1])
        {
            /* perform for coords */
            case 'c':
                i++;
                call_crane = 1;
                
                if(i>argc)
                  continue;
                
                x = atol(strtok(argv[i], ","));
                x *= (long int)(xtime);

                /* white space or none? */
                if((i+1<argc) && IS_NUM(argv[i+1][0]))
                  y = atof(argv[i+1]);
                else
                  y = atol(strtok(NULL, "\n"));

                y *= (long int)(ytime);
                break;

            /* demo */
            case 'd':
                demo_crane();
                break;

            /* email */
            case 'e':
                i++;
                strncpy(email, argv[i], sizeof(email));
                break;
            
            /* hash */
            case 'H':
                i++;
                strncpy(hash, argv[i], sizeof(hash));
                break;
            
            /* initalize */
            case 'i':
                init_crane();
                break;
            
            /* no pics */
            case 'n':
                is_take_pics = 0;
                break;

            /* pic path */
            case 'p':
                i++;
                strncpy(pic_path, argv[i], sizeof(pic_path));
                break;
            
            /* quiet/verbose */
            case 'q':
                is_verbose = 0;
                break;
            
            /* screwed up */
            case 'h':
            default:
                usage();
                exit(-1);
        }
    }

    /* Post argv processin */
    if(call_crane)
    {
        /* Format for picture taking */
        snprintf(script_args, sizeof(script_args), 
            "%s %s %s", pic_path, email, hash);
#ifdef DEBUG
        printf("DEBUG: %s\n", script_args);
#endif DEBUG
        
        cranage(x, y, script_args);
    }
    
    return 0;
}
