#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <asm/io.h>
#include <time.h>
#include <sys/time.h>

/* Defaults */
#undef  DEBUG
#define PORT     0x378
#define PORT_IN  PORT+1
#define VERSION  "0.03"
#define APP_NAME "Crane"
#define BIN_NAME "crane"
#define DEFAULT_CONFIG_FILE ".crane.cfg"

/* Macs */
#define IS_NUM(x) ( ((int)x)>47 & ((int)x<58) )
#define CLAW_SHIFT(x, c) ( (c) ? x+1 : x )
#define TIMES_TO_TICKS(t) (t/100)

/* Flags and such */
static char g_config_file[50] = DEFAULT_CONFIG_FILE;
static int is_verbose     = 1;
static long int xtime     = 0;
static long int ytime     = 0;

/* Demo */
static enum {STROKE, SWIRL} d_mode;

/* Pins */
#define MV_LEFT       24
#define MV_RIGHT      16
#define MV_FRONT      64
#define MV_BACK       96
#define MV_CLAW_UP    6
#define MV_CLAW_DOWN  4
#define MV_CLAW_CLOSE 1
static int pin_status[4];
static int pin_fb_home   = 0;
static int pin_lr_home   = 1;
static int pin_claw_up   = 2;
static int pin_claw_down = 3;


/* Move by clock ticks */
void move_tick(
	int direction,
	int status_idx,
	int ticks,
	int force)
{
	int i;
    clock_t ct_a;

	if(pin_status[status_idx] || force)
    {
        ct_a = clock();
        for(i=0; i<ticks; i++)
        {
            outb(direction, PORT);
            if(clock()-ct_a > ticks)
              break;
        }
	    outb(0, PORT);
    }
}


/* Duh */
void claw_fetch(void)
{
    /* Drop */
    while(!pin_status[pin_claw_down])
    {
        get_input();
        outb(MV_CLAW_DOWN, PORT);
    }
    usleep(200);
    outb(0, PORT);
    sleep(1);

    /* Up */
    while(!pin_status[pin_claw_up])
    {
        get_input();
        outb(CLAW_SHIFT(MV_CLAW_UP, 1), PORT);
    }
    sleep(1);
}


/* What the inputs are */
void print_pins(void)
{
	int i;

	for(i=0; i<4; i++)
	  printf("pins[%d]: %d\n", i, pin_status[i]);
}


/* What is the crane tellin us? 
 * Return high 4 bits representing HOME and CLAW status 
 */
int get_input(void)
{
        unsigned char ret;
        int i;
        
        ret = inb(PORT_IN);
        ret >>= 4;

        for(i=0; i<4; i++)
          pin_status[i] = (ret>>i)%2;

        return ret;
}


/* Go to the crane's origin (Where the prizes are dropped) */
void go_origin(int is_claw)
{
    int is_good = 0;
    int output;

    /* Is the claw grabbin some booty */
    
    /* Dont ask */
    move_tick(CLAW_SHIFT(MV_LEFT, is_claw), 
              pin_lr_home, 100000, 1);
	move_tick(CLAW_SHIFT(MV_FRONT, is_claw), 
              pin_fb_home, 100000, 1);

    get_input();
    while(!is_good)
    {
        output = 0;

        /* Bring claw up */
        while(!pin_status[pin_claw_up])
        {
            outb(CLAW_SHIFT(MV_CLAW_UP, is_claw), PORT);
            get_input();
            
            /* Stop yo */
            if(pin_status[pin_claw_up])
              outb(0, PORT);
        }

        /* Claw working */
        output = CLAW_SHIFT(output, is_claw);
        
        /* Left right home */
        if(pin_status[pin_lr_home])
          output += MV_LEFT;
        
        /* Front back home */
        if(pin_status[pin_fb_home])
          output += MV_FRONT;

        /* Go */
        outb(output, PORT);

        if(is_claw)
          output--;

        if(output == 0)
          is_good = 1;

        /* Update */
        get_input();
		
#ifdef DEBUG
		print_pins();
#endif
    }
}


/* Go to the point and get some prizes */
void cranage(
    double x, 
    double y)
{
    /* go_point() */
    move_tick(MV_RIGHT, 
              pin_lr_home, x, 1);
    
    move_tick(MV_BACK, 
              pin_fb_home, y, 1);
    
    claw_fetch();
    go_origin(1);
    outb(0, PORT);
}


/* Load configs */
void load_configs(char *fname)
{
    FILE *fp;
    char line[64];
    
    memset(line, '\0', sizeof(line));
    if(fp = fopen(fname, "r"))
    {
        /* Skip comments */
        while(fgets(line, 32, fp) && line[0] == ';');
     
        /* Tick times */
        xtime = TIMES_TO_TICKS(atol(strtok(line, ",")));
        ytime = TIMES_TO_TICKS(atol(strtok(NULL, ",")));
#ifdef DEBUG
        printf("loading xtime(%i) ytime(%i)\n", xtime, ytime);
#endif
        fclose(fp);
    }
    else /* dont die */
      printf("Error opening config file to load: %s\n",
              fname);
}


/* Save configs */
void save_configs(
    char *fname,
    char *data)
{
    FILE *fp;
    
    if(!(fp = fopen(fname, "w")))
      printf("Error opening config file to save: %s\n",
              fname);
    printf("data: %s\n", data);
    fprintf(fp, data, sizeof(char) * (strlen(data) + 1));
    fclose(fp);
}
        

/* Initalize Crane */
void init_crane(void)
{
    char save_data[256];
    char str[50];
	clock_t ct_a, ct_b;
    
    sprintf(save_data, "; Config created: %d\n", time(NULL));

    if(is_verbose)
      printf("Initalizing crane...\n");

    /* Go to your home */
    go_origin(0);

	/* Callibrate X-axis */
	pin_status[pin_lr_home] = 1;
	ct_a = clock();
	outb(MV_RIGHT, PORT);
	usleep(100000);
	while(pin_status[pin_lr_home])
  	  get_input();
	ct_b = clock();
	outb(0, PORT);
	xtime = ct_b-ct_a;

	if(is_verbose)
	  printf("X-Axis time: %d\n", xtime);
	
	/* Callibrate Y-axis */
	pin_status[pin_fb_home] = 1;
	ct_a = clock();
	outb(MV_BACK, PORT);
	usleep(100000);
	while(pin_status[pin_fb_home])
  	  get_input();
	ct_b = clock();
	outb(0, PORT);
	ytime = ct_b-ct_a;

	if(is_verbose)
	  printf("Y-Axis time: %d\n", ytime);

    /* Save to config file */
    sprintf(str, "%d,%d\n", xtime, ytime);
    strcat(save_data, str);
    save_configs(g_config_file, save_data);

    /* Return */
    go_origin(0);
}


/* Demo Crane */
void demo_crane(void)
{
    int i;
    
    if(is_verbose)
      printf("Demoing crane...\n");

    /* Do some laps, have it lower claw,
     * and swing the claw so it taps the glass
     * possibly go for a pendulum effect
     * consider disco ball
     */

    /* Stroke */
    d_mode = STROKE;
    switch(d_mode)
    {
        case STROKE:
            go_origin(0);
            move_tick(MV_RIGHT, pin_lr_home, xtime/2, 1);
            move_tick(MV_BACK, pin_lr_home, ytime/2, 1);
            for(i=0; i<(1+rand())%10; i++)
            {
                outb(MV_CLAW_DOWN, PORT);
                usleep(600000);
                get_input();
                while(!pin_status[pin_claw_up])
                {
                    get_input();
                    outb(CLAW_SHIFT(MV_CLAW_UP, 1), PORT);
                }
                usleep(300000);
            }
            outb(0, PORT);
            go_origin(0);
            break;
    }
}


/* Usage */
void usage(void)
{
    printf("--{ %s -- v%s }--\n"
           "Usage: ./%s [-c <x,y>] [-d] [-i] [-v] ... [-h]\n"
           "\t-c <x, y> Attack Coordinate (X, Y)\n"
           "\t-d Demo Mode\n"
           "\t-i Initalize Crane\n"
           "\t-h Help\n"
           "\t-q Quiet\n",
           APP_NAME, VERSION, BIN_NAME);
}


/* Main */
int main(
    int    argc,
    char **argv)
{
    int i;
    long int x, y;

    srand(time(NULL));

    ioperm(PORT_IN, 1, 1);
    ioperm(PORT, 1, 1);
    
    /* Argz:
     * 
     * -i initalize
     * -h help
     * -c <int, int>
     * -d demo
     * -q quiet
     */
    if(argc == 1)
    {
        usage();
        _Exit(-1);
    }
    else
      load_configs(g_config_file);

    for(i=1; i<argc; i++)
    {
        if(argv[i][0] != '-')
          continue;
          
        switch(argv[i][1])
        {
            /* perform for coords */
            case 'c':
                i++;
                if(i>argc)
                  continue;
                
                x = atol(strtok(argv[i], ","));
                x *= (long int)(xtime);

                /* white space or none? */
                if((i+1<argc) && IS_NUM(argv[i+1][0]))
                  y = atof(argv[i+1]);
                else
                  y = atol(strtok(NULL, "\n"));

                y *= (long int)(ytime);
                printf("%d,%d\n", x, y);
                cranage(x, y);
                break;

            /* demo */
            case 'd':
                demo_crane();
                break;
            
            /* initalize */
            case 'i':
                init_crane();
                break;
            
            /* quiet/verbose */
            case 'q':
                is_verbose = 0;
                break;
            
            /* screwed up */
            case 'h':
            default:
                usage();
                _Exit(-1);
        }
    }
    
    return 0;
}
